/*
 * Copyright 2009 Arne Vajhj.
 * 
 * This code is available under either Apache 2.0 license:
 *     http://www.apache.org/licenses/LICENSE-2.0
 * or AGPL version 3 license:
 *     https://www.gnu.org/licenses/agpl-3.0.html 
 *     
 * Berkeley DB (BDB) is dual licensed under AGPL and a commercial license from Oracle.
 * 
 * I strongly recommend seeking legal advice on implications from these licenses
 * before using BDB. But this is my understanding as a non-lawyer.
 * 
 * If you have a commercial license from Oracle then you can take this code under Apache license
 * and use whatever license you want for your own code.
 * 
 * If you have taken BDB under AGPL then you need to take this code under AGPL license and
 * also use AGPL for your own code.
 * 
 * The rest of the "isam" and "record" packages are not linked with BDB code and can be used under
 * Apache license no matter what license you have taken BDB under.
 *
 */
package dk.vajhoej.bdb;

import java.io.FileNotFoundException;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sleepycat.db.Cursor;
import com.sleepycat.db.Database;
import com.sleepycat.db.DatabaseConfig;
import com.sleepycat.db.DatabaseEntry;
import com.sleepycat.db.DatabaseException;
import com.sleepycat.db.OperationStatus;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.LogHelper;
import dk.vajhoej.isam.local.NativeIsam;

public class BDB implements NativeIsam {
	private static Logger log = Logger.getLogger(BDB.class.getName());
	private static byte[] i2b(int v) {
		byte[] res = new byte[4];
		for(int i = 0; i < 4; i++) {
			res[i] = (byte)v;
			v >>= 8;
		}
		return res;
	}
	private static DatabaseEntry prep(Object k) throws IsamException {
		if(k instanceof byte[]) {
			return new DatabaseEntry((byte[])k);
		} else if(k instanceof String) {
			return new DatabaseEntry(((String)k).getBytes());
		} else if(k instanceof Integer) {
			return new DatabaseEntry(i2b(((Integer)k).intValue()));
		} else {
			throw new IsamException(k.getClass().getName() + " is not a supported key");
		}
	}
	private Database db;
	private Cursor it;
	public void open(String fnm, boolean readonly) throws IsamException {
        log.fine("File open " + fnm);
		DatabaseConfig cfg = new DatabaseConfig();
		cfg.setReadOnly(readonly);
		try {
			db = new Database(fnm, null, cfg);
            log.fine("Success");
		} catch (FileNotFoundException e) {
            log.fine("Error");
			throw new IsamException(e);
		} catch (DatabaseException e) {
            log.fine("Error");
			throw new IsamException(e);
		}
	}
	public int read(int ix, Object k, byte[] buf) throws IsamException {
		if(ix != 0) {
			throw new IsamException("Index <> 0 is not supported");
		}
		DatabaseEntry key = prep(k);
		DatabaseEntry value = new DatabaseEntry();
		try {
	        log.fine("Reading key " + ix + " " + k + " from file " + db.getDatabaseFile());
			OperationStatus stat = db.get(null,  key,  value, null);
	       if(log.isLoggable(Level.FINEST)) {
	            log.finest("Status: " + stat);
	            log.finest("Bytes:" + LogHelper.byteArrayToString(value.getData(), 16));
	        }
			if(stat != OperationStatus.SUCCESS) return -1;
			System.arraycopy(value.getData(), 0, buf, 0, value.getSize());
			return value.getSize();
		} catch (DatabaseException e) {
			throw new IsamException(e);
		}
	}
	public int create(byte[] buf) throws IsamException {
		throw new IsamException("Not possible to create without key");
	}
	public int create(byte[] buf, int ix, Object k) throws IsamException {
		try {
	        log.fine("Creating in file " + db.getDatabaseFile());
	        if(log.isLoggable(Level.FINEST)) {
	            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
	        }
			OperationStatus stat = db.put(null, prep(k), new DatabaseEntry(buf));
			return stat == OperationStatus.SUCCESS ? buf.length : -1;
		} catch (DatabaseException e) {
			throw new IsamException(e);
		} catch (IsamException e) {
			throw new IsamException(e);
		}
	}
	public int update(byte[] buf) throws IsamException {
		throw new IsamException("Not possible to update without key");
	}
	public int update(byte[] buf, int ix, Object k) throws IsamException {
		try {
	        log.fine("Updating in file " + db.getDatabaseFile());
	        if(log.isLoggable(Level.FINEST)) {
	            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
	        }
			OperationStatus stat = db.put(null, prep(k), new DatabaseEntry(buf));
			return stat == OperationStatus.SUCCESS ? buf.length : -1;
		} catch (DatabaseException e) {
			throw new IsamException(e);
		} catch (IsamException e) {
			throw new IsamException(e);
		}
	}
	public int delete(int ix, Object k) throws IsamException {
		try {
	        log.fine("Deleting key " + ix + " " + k + " from file " + db.getDatabaseFile());
			OperationStatus stat = db.delete(null, prep(k));
			return stat == OperationStatus.SUCCESS ? 1 : -1;
		} catch (DatabaseException e) {
			throw new IsamException(e);
		} catch (IsamException e) {
			throw new IsamException(e);
		}
	}
	public int find(int ix, Object k, int comp, byte[] buf) throws IsamException {
		if(comp != COMP_GE) {
			throw new IsamException(comp + " comparison not supported");
		}
		DatabaseEntry key = prep(k);
		DatabaseEntry value = new DatabaseEntry();
		try {
	        log.fine("Finding key " + ix + " " + k + " in file " + db.getDatabaseFile());
			it = db.openCursor(null, null);
			OperationStatus stat = it.getSearchKeyRange(key, value, null);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Status: " + stat);
                log.finest("Bytes:" + LogHelper.byteArrayToString(value.getData(), 16));
            }
			if(stat != OperationStatus.SUCCESS) return -1;
			System.arraycopy(value.getData(), 0, buf, 0, value.getSize());
			return value.getSize();
		} catch (DatabaseException e) {
			throw new IsamException(e);
		}
	}
	public int findnext(byte[] buf) throws IsamException {
		DatabaseEntry key = new DatabaseEntry();
		DatabaseEntry value = new DatabaseEntry();
		try {
	        log.fine("Find next in file " + db.getDatabaseFile());
			OperationStatus stat = it.getNext(key, value, null);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Status: " + stat);
                log.finest("Bytes:" + LogHelper.byteArrayToString(value.getData(), 16));
            }
			if(stat != OperationStatus.SUCCESS) return -1;
			System.arraycopy(value.getData(), 0, buf, 0, value.getSize());
			return value.getSize();
		} catch (DatabaseException e) {
			throw new IsamException(e);
		}
	}
	public void close() throws IsamException {
		try {
	        log.fine("Close file " + db.getDatabaseFile());
			if(it != null) it.close();
			db.close();
		} catch (DatabaseException e) {
			throw new IsamException(e);
		}
	}
	public boolean supportReverse() {
		return false;
	}
	public boolean supportModWithoutKey() {
		return false; // needed for API
	}
    public boolean supportFindMixedOther() {
    	return true; // thanks to cursor
    }
}
