/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.local;

import java.lang.reflect.Field;
import java.util.logging.Level;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.isam.Key;
import dk.vajhoej.isam.KeyInfo;
import dk.vajhoej.isam.KeyInfoArrayCache;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.StructInfo;
import dk.vajhoej.record.StructInfoCache;
import dk.vajhoej.record.StructReader;
import dk.vajhoej.record.StructWriter;

/**
 * Class LocalIsamSource is accessing a local ISAM file using native API.
 */
public class LocalIsamSource implements IsamSource {
    private static Logger log = Logger.getLogger(LocalIsamSource.class.getName());
    private String fnm;
    private String implementation;
    private NativeIsam del;
    /**
     * Create instance of LocaleIsamSource.
     * @param fnm filename
     * @param implementation ISAM implementation class
     * @param readonly true=&gt;only read, false=&gt;read+create+update+delete
     * @throws IsamException if error creating instance of ISAM implementation class
     */
	public LocalIsamSource(String fnm, String implementation, boolean readonly) throws IsamException {
	    this.fnm = fnm;
	    this.implementation = implementation;
	    try {
            del = (NativeIsam)Class.forName(implementation).newInstance();
        } catch (InstantiationException e) {
            throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
        } catch (IllegalAccessException e) {
            throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Error creating native ISAM implementation - class not found: " + implementation);
        }
		del.open(fnm, readonly);
        log.fine(fnm + " " + implementation + " initialized");
	}
	public <T, TK> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Read class " + t.getName() + " with key " + k);
        StructInfo si = StructInfoCache.analyze(t);
        int len = si.getLength();
        if(!si.isFixedLength()) {
            len += 32768;
        }
        log.finer("Reading with buffer length " + len);
        byte[] buf = new byte[len];
        int stat = del.read(k.getIx(), k.getValue(), buf);
	    if(stat <= 0) {
            log.fine("Not found");
	        return null;
	    } else {
            log.fine("Found");
	    }
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
        }
        StructReader sr = new StructReader(buf);
        return sr.read(t);
	}
	public <T> void create(T o) throws IsamException, RecordException {
        log.fine("Create class " + o.getClass().getName() + " with value " + o);
        StructWriter sw = new StructWriter();
        sw.write(o);
        byte[] buf = sw.getBytes();
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        int stat;
        if(del.supportModWithoutKey()) {
        	stat = del.create(buf);
        } else {
            log.finer("Key needed");
            KeyInfo[] keys = KeyInfoArrayCache.analyze(o.getClass());
            Field f = keys[0].getField();
            f.setAccessible(true);
            try {
                Object k = f.get(o);
                log.fine("Key found " + k);
                stat = del.create(buf, 0, k);
            } catch (IllegalArgumentException e) {
                throw new IsamException("Problem creating");
            } catch (IllegalAccessException e) {
                throw new IsamException("Problem creating");
            }
        }
        if(stat != buf.length) {
            log.fine("Failure");
            throw new IsamException("Error creating record in file");
        } else {
            log.fine("Success");
        }
	}
    public <T> void update(T o) throws IsamException, RecordException {
        log.fine("Update class " + o.getClass().getName() + " with value " + o);
        StructWriter sw = new StructWriter();
        sw.write(o);
        byte[] buf = sw.getBytes();
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        int stat;
        if(del.supportModWithoutKey()) {
            stat = del.update(buf);
        } else {
            log.finer("Key needed");
            KeyInfo[] keys = KeyInfoArrayCache.analyze(o.getClass());
            Field f = keys[0].getField();
            f.setAccessible(true);
            try {
                Object k = f.get(o);
                log.fine("Key found " + k);
                stat = del.update(buf, 0, k);
            } catch (IllegalArgumentException e) {
                throw new IsamException("Problem updating");
            } catch (IllegalAccessException e) {
                throw new IsamException("Problem updating");
            }
        }
        if(stat != buf.length) {
            log.fine("Failure");
            throw new IsamException("Error updating record in file");
        } else {
            log.fine("Success");
        }
    }
    public <T, TK> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Delete class " + t.getName() + " with key " + k);
        int stat = del.delete(k.getIx(), k.getValue());
        if(stat <= 0) {
            log.fine("Failure");
            throw new IsamException("Error deleting record from file");
        } else {
            log.fine("Success");
        }
    }
	public void close() throws IsamException {
	    del.close();
        log.fine("Closed");
	}
	public <T, TK> IsamResult<T> readGE(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadGE class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_GE);
	}
	public <T, TK> IsamResult<T> readGT(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadGT class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_GT);
	}
	public <T, TK> IsamResult<T> readLE(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadLE class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_LE);
	}
	public <T, TK> IsamResult<T> readLT(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadLT class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_LT);
	}
	private <T, TK> IsamResult<T> readXX(Class<T> t, Key<TK> k, int comp) throws IsamException, RecordException {
        NativeIsam del2;
        if(!del.supportFindMixedOther()) {
            try {
                del2 = (NativeIsam)Class.forName(implementation).newInstance();
            } catch (InstantiationException e) {
                throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
            } catch (IllegalAccessException e) {
                throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
            } catch (ClassNotFoundException e) {
                throw new IsamException("Error creating native ISAM implementation - class not found: " + implementation);
            }
            del2.open(fnm, true);
            log.fine(fnm + " " + implementation + " initialized");
        } else {
        	del2 = del;
        }
        StructInfo si = StructInfoCache.analyze(t);
        int len = si.getLength();
        if(!si.isFixedLength()) {
            len += 32768;
        }
        log.finer("Finding with buffer length " + len);
        byte[] buf = new byte[len];
        int stat = del2.find(k.getIx(), k.getValue(), comp, buf);
	    if(comp == NativeIsam.COMP_GE || comp == NativeIsam.COMP_GT || del.supportReverse()) {
            return new LocalIsamResult<T>(t, del2, buf, stat > 0);
    	} else {
            return new LocalIsamResultReverse<T, TK>(t, k, del2, buf, stat > 0);
    	}	    
	}
}
