/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.rms;

import java.util.logging.Level;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.LogHelper;
import dk.vajhoej.isam.local.NativeIsam;

/**
 * Class IndexSequential is the implementation of ISAM for OpenVMS RMS.
 */
public class IndexSequential implements NativeIsam {
    private static Logger log = Logger.getLogger(IndexSequential.class.getName());
    private long ptr;
    public void open(String fnm, boolean readonly) throws IsamException {
        log.fine("File open " + fnm);
        ptr = NativeAPI.open(fnm, readonly);
        if(ptr == 0) {
            log.fine("Error");
            throw new IsamException("Error opening file: " + fnm);
        } else {
            log.fine("Returning file pointer " + ptr);
        }
    }
    public int read(int ix, Object k, byte[] buf) throws IsamException {
        log.fine("Reading key " + ix + " " + k + " from file " + ptr);
        int stat;
        Class<?> clz = k.getClass();
        if(clz.equals(Byte.class)) {
            stat = NativeAPI.getByte(ptr, (byte)ix, (Byte)k, buf);
        } else if(clz.equals(Short.class)) {
            stat = NativeAPI.getShort(ptr, (byte)ix, (Short)k, buf);
        } else if(clz.equals(Integer.class)) {
            stat = NativeAPI.getInt(ptr, (byte)ix, (Integer)k, buf);
        } else if(clz.equals(Long.class)) {
            stat = NativeAPI.getLong(ptr, (byte)ix, (Long)k, buf);
        } else if(clz.equals(String.class)) {
            stat = NativeAPI.getString(ptr, (byte)ix, (String)k, buf);
        } else {
            throw new IsamException(clz.getName() + " can not be used as a key");
        }
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Status: " + stat);
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
        }
        return stat;
    }
    public int create(byte[] buf) throws IsamException {
        log.fine("Creating in file " + ptr);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        return NativeAPI.put(ptr, buf);
    }
    public int create(byte[] buf, int ix, Object k) throws IsamException {
    	// no special action needed
    	return create(buf);
    }
    public int update(byte[] buf) throws IsamException {
        log.fine("Updating in file " + ptr);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        return NativeAPI.update(ptr, buf);
    }
    public int update(byte[] buf, int ix, Object k) throws IsamException {
        log.fine("Updating in file " + ptr);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        // need to lock key
        Class<?> clz = k.getClass();
        if(clz.equals(Byte.class)) {
            NativeAPI.lockByte(ptr, (byte)ix, (Byte)k);
        } else if(clz.equals(Short.class)) {
            NativeAPI.lockShort(ptr, (byte)ix, (Short)k);
        } else if(clz.equals(Integer.class)) {
            NativeAPI.lockInt(ptr, (byte)ix, (Integer)k);
        } else if(clz.equals(Integer.class)) {
            NativeAPI.lockLong(ptr, (byte)ix, (Long)k);
        } else if(clz.equals(String.class)) {
            NativeAPI.lockString(ptr, (byte)ix, (String)k);
        } else {
            throw new IsamException(clz.getName() + " can not be used as a key");
        }
        return NativeAPI.update(ptr, buf);
    }
    public int delete(int ix, Object k) throws IsamException {
        log.fine("Deleting key " + ix + " " + k + " from file " + ptr);
        Class<?> clz = k.getClass();
        if(clz.equals(Byte.class)) {
            NativeAPI.lockByte(ptr, (byte)ix, (Byte)k);
            return NativeAPI.deleteByte(ptr, (byte)ix, (Byte)k);
        } else if(clz.equals(Short.class)) {
            NativeAPI.lockShort(ptr, (byte)ix, (Short)k);
            return NativeAPI.deleteShort(ptr, (byte)ix, (Short)k);
        } else if(clz.equals(Integer.class)) {
            NativeAPI.lockInt(ptr, (byte)ix, (Integer)k);
            return NativeAPI.deleteInt(ptr, (byte)ix, (Integer)k);
        } else if(clz.equals(Long.class)) {
            NativeAPI.lockLong(ptr, (byte)ix, (Long)k);
            return NativeAPI.deleteInt(ptr, (byte)ix, (Integer)k);
        } else if(clz.equals(String.class)) {
            NativeAPI.lockString(ptr, (byte)ix, (String)k);
            return NativeAPI.deleteString(ptr, (byte)ix, (String)k);
        } else {
            throw new IsamException(clz.getName() + " can not be used as a key");
        }
    }
    public int find(int ix, Object k, int comp, byte[] buf) throws IsamException {
        log.fine("Finding key " + ix + " " + k + " in file " + ptr);
        Class<?> clz = k.getClass();
        int fnd;
        if(clz.equals(Byte.class)) {
            fnd = NativeAPI.findByte(ptr, (byte)ix, (Byte)k, comp);
        } else if(clz.equals(Short.class)) {
            fnd = NativeAPI.findShort(ptr, (byte)ix, (Short)k, comp);
        } else if(clz.equals(Integer.class)) {
            fnd = NativeAPI.findInt(ptr, (byte)ix, (Integer)k, comp);
        } else if(clz.equals(Long.class)) {
            fnd = NativeAPI.findLong(ptr, (byte)ix, (Long)k, comp);
        } else if(clz.equals(String.class)) {
            fnd = NativeAPI.findString(ptr, (byte)ix, (String)k, comp);
        } else {
            throw new IsamException(clz.getName() + " can not be used as a key");
        }
        if(fnd > 0) {
            int stat = NativeAPI.get(ptr, buf);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Status: " + stat);
                log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
            }
            return stat;
        } else {
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Status: " + fnd);
                log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
            }
            return fnd;
        }
    }
    public int findnext(byte[] buf) throws IsamException {
        log.fine("Find next in file " + ptr);
        int stat = NativeAPI.get(ptr, buf);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Status: " + stat);
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
        }
        return stat;
    }
    public void close() throws IsamException {
        log.fine("Close file " + ptr);
        NativeAPI.close(ptr);
        ptr = 0;
    }
    public boolean supportReverse() {
        return false;
    }
    public boolean supportModWithoutKey() {
        return false; // needed for locking
    }
    public boolean supportFindMixedOther() {
    	return false;
    }
}
